/* Precomp.h -- StdAfx
2023-03-04 : Igor Pavlov : Public domain */

#ifndef ZIP7_INC_PRECOMP_H
#define ZIP7_INC_PRECOMP_H

#if defined(_MSC_VER) && _MSC_VER >= 1800
#pragma warning(disable : 4464) // relative include path contains '..'
#endif

#include "../../Compiler.h"
#include "../../7zTypes.h"

#endif
